import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import ij.plugin.*;
import ij.plugin.MeasurementsWriter;
import ij.util.Tools;
import ij.measure.*;
import ij.text.*;
import java.lang.Math;

public class Detection_Pairing implements PlugIn {

	ImagePlus[] image = new ImagePlus [2];
	ImageProcessor[] imp = new ImageProcessor [2];

	int[] Ax;
	int[] Ay;
	int[] Bx;
	int[] By;

	int numA, numB;

	public class Detection_Matrix {
		public int rows;
		public int columns;

		double[] mtx;

		Detection_Matrix( int r, int c )
		{
			mtx = new double [ r * c ];
			rows = r;
			columns = c;
		}

		double getElement(int r, int c)
		{
			return mtx [ r * columns + c];
		}
		
		void setElement(int r, int c, double value)
		{
			mtx [ r * columns + c] = value;
		}

		void show3x4()
		{
			IJ.showMessage(mtx[0] + " " + mtx[1] + " " + mtx[2] + " " + mtx[3] + "\n" + mtx[4] + " " + mtx[5] + " " + mtx[6] + " " + mtx[7] + "\n" + mtx[8] + " " + mtx[9] + " " + mtx[10] + " " + mtx[11]);
		}

		void show()
		{
			String str = new String();
			
			for( int i = 0; i < rows; i++ )
			{
				for( int j = 0; j < columns; j++ )
				{
					str += getElement( i, j );
					str += " ";
				}

				str += "\n";
			}

			IJ.showMessage( str );
		}
		
	}

	private int Jordan ( Detection_Matrix s )
	{
		int i, j, k;
		double x;
		double value;

		for(i=0; i<s.rows; i++)
		{
			x = s.getElement( i, i );
			if( x == 0.0)
			{
				for( j = i + 1; j < s.rows; j++ )
				if(s.getElement( i, i ) != 0.0)
					break;

				if ( j == s.rows )
					return 0;

			for( k = 0; k < s.columns; k++)
			{
				x = s.getElement( i, k );
				value = s.getElement( j, k );
				s.setElement( i, k, value );
				s.setElement( j, k, x );
			}

			x = s.getElement(i, i);
		}

		s.setElement( i, i, 1.0 );
		for( j = i + 1; j < s.columns; j++)
		{
			value = s.getElement( i, j ) / x;
			s.setElement( i, j, value );
		}

		for( j = 0; j < s.rows; j++)
		{
			if( j != i)
			{
				x = s.getElement( j, i );
				s.setElement( j, i, 0.0 );
				for( k = i + 1; k < s.columns; k++)
				{
					value = s.getElement( j, k ) - x * s.getElement( i, k );
					s.setElement( j, k, value );
				}
			}
		}
	}



		return 1;
	}


	public boolean imagesDialog()
	{
		int[] wList = WindowManager.getIDList();
		if ( ( wList==null ) || ( wList.length < 2 ) ) {
			IJ.error("At least two images required.");
			return false;
		}

		String[] titles = new String[wList.length];
		for (int i=0; i<wList.length; i++) {
			ImagePlus imp = WindowManager.getImage(wList[i]);
			titles[i] = imp!=null?imp.getTitle():"";
		}

		GenericDialog gd = new GenericDialog("PointSelection Test");
		gd.addChoice("Reference:", titles, titles[0]);
		gd.addChoice("Moving:", titles, titles[1]);

		gd.showDialog();
		if (gd.wasCanceled()) 
			return false;
		for(int ii=0;ii<2;ii++) {
			int FILE=gd.getNextChoiceIndex();
			image[ii] = (FILE<wList.length)
					? WindowManager.getImage(wList[FILE])
					: null;

			imp[ii] = ( image[ii] != null)
					? image[ii].getProcessor()
					: null;
		}

		return true;
	}

	private boolean typeCheck()
	{
		int type;

		type = image[0].getType();

		if ( type != ImagePlus.GRAY8 )
		{
			IJ.error("Reference image must be 8-bit grayscale");
			return false;
		}

		type = image[1].getType();

		if ( type != ImagePlus.GRAY8 )
		{
			IJ.error("Moving image must be 8-bit template grayscale");
			return false;
		}

		return true;
	}

	private boolean getCoordinates()
	{
		int i;

		Roi refroi = image[0].getRoi();
		int roiType = refroi.getType();

		if ( roiType != 10 )
		{
			IJ.error("Point selection required!");
			return false;
		}

		Rectangle refrect = refroi.getBounds();
		int refbx = refrect.x;
		int refby = refrect.y;

		PolygonRoi refp = ( PolygonRoi ) refroi;
	            numA = refp.getNCoordinates();
	            Ax = refp.getXCoordinates();
	            Ay = refp.getYCoordinates();

/*
		for ( i = 0; i < numA; i++ )
		{
			Ax[ i ] += refbx;
			Ay[ i ] += refby;
		}
*/

//

		Roi movroi = image[1].getRoi();
		roiType = movroi.getType();

		if ( roiType != 10 )
		{
			IJ.error("Point selection required!");
			return false;
		}

		Rectangle movrect = movroi.getBounds();
		int movbx = movrect.x;
		int movby = movrect.y;

		PolygonRoi movp = ( PolygonRoi ) movroi;
	            numB = movp.getNCoordinates();
	            Bx = movp.getXCoordinates();
	            By = movp.getYCoordinates();

/*
		for ( i = 0; i < numB; i++ )
		{
			Bx[ i ] += movbx;
			By[ i ] += movby;
		}
*/

		return true;
	}


	public void run(String arg) {
		int i, j, k, i1, i2, i3, i4, maxc;
		int j1, j2, j3, itrf, v;
		int max_i1, max_i2, max_i3;
		double tAx, tAy, tBx, tBy, tCx, tCy;
		double oAx, oAy, oBx, oBy, oCx, oCy;
		Detection_Matrix s;
		double max_t11, max_t12, max_t13, max_t21, max_t22, max_t23;
		double mean, min_mean;
		double mindiff;
		int mindiffidx;

		if( imagesDialog() == false )
			return;

		if( typeCheck() == false )
			return;

		if ( getCoordinates() == false )
			return;


		int[] pairsA = new int [ numA ];
		int[] pairsA_temp = new int [ numA ];

		double[] pairsA_diff = new double [ numA ];

		int[] pairsB = new int [ numB ];
		int[] pairsB_temp = new int [ numB ];


		max_i1 = -1;
		max_i2 = -1;
		max_i3 = -1;

		max_t11 = max_t12 = max_t13 = max_t21 = max_t22 = max_t23 = 0.0;
		min_mean = 1000000.0;

		s = new Detection_Matrix( 3, 4 );

		int count = 0;
		double t11, t12, t13, t21, t22, t23;
		int valid = 0;
		int invalid = 0;
		int maxvalid = 0;
		int novalid = 1;

		for( i1 = 0; i1 < numA - 2; i1++ )
			for( i2 = 1; i2 < numA - 1; i2++ )
				for( i3 = 2; i3 < numA; i3++ )
		for( j1 = 0; j1 < numB - 2; j1++ )
			for( j2 = 1; j2 < numB - 1; j2++ )
				for( j3 = 2; j3 < numB; j3++ )
				{
			s.setElement( 0, 0, Ax[i1] );
			s.setElement( 0, 1, Ay[i1] );
			s.setElement( 0, 2, 1.0 );
			s.setElement( 0, 3, Bx[j1] );

			s.setElement( 1, 0, Ax[i2] );
			s.setElement( 1, 1, Ay[i2] );
			s.setElement( 1, 2, 1.0 );
			s.setElement( 1, 3, Bx[j2] );

			s.setElement( 2, 0, Ax[i3] );
			s.setElement( 2, 1, Ay[i3] );
			s.setElement( 2, 2, 1.0 );
			s.setElement( 2, 3, Bx[j3] );

			Jordan( s );

			t11 = s.getElement( 0, 3 );
			t12 = s.getElement( 1, 3 );
			t13 = s.getElement( 2, 3 );

//

			s.setElement( 0, 0, Ax[i1] );
			s.setElement( 0, 1, Ay[i1] );
			s.setElement( 0, 2, 1.0 );
			s.setElement( 0, 3, By[j1] );

			s.setElement( 1, 0, Ax[i2] );
			s.setElement( 1, 1, Ay[i2] );
			s.setElement( 1, 2, 1.0 );
			s.setElement( 1, 3, By[j2] );

			s.setElement( 2, 0, Ax[i3] );
			s.setElement( 2, 1, Ay[i3] );
			s.setElement( 2, 2, 1.0 );
			s.setElement( 2, 3, By[j3] );

			Jordan( s );

			t21 = s.getElement( 0, 3 );
			t22 = s.getElement( 1, 3 );
			t23 = s.getElement( 2, 3 );

//  Execute transformation

			double trfx, trfy, diff;
			valid = 0;
			invalid = 0;
			mean = 0.0;

			for( j = 0; j < numB; j++ )
				pairsB_temp[ j ] = -1;

			pairsA_temp[ i1 ] = j1;
			pairsA_temp[ i2 ] = j2;
			pairsA_temp[ i3 ] = j3;

			for( itrf = 0; itrf < numA; itrf++ )
			{
				if( ( itrf != i1 ) && ( itrf != i2 ) && ( itrf != i3 ) )
				{
					trfx = ( t11 * Ax[itrf] + t12 * Ay[itrf] + t13 );
					trfy = ( t21 * Ax[itrf] + t22 * Ay[itrf] + t23 );

					diff = mindiff = 10000000.0;
					mindiffidx = -1;
					for( v = 0; v < numB; v++ )
					{
						if( ( v != j1 ) && ( v != j2 ) && ( v != j3 ) )
						{
							diff = Math.sqrt( ( Bx[v] - trfx ) * ( Bx[v] - trfx ) + ( By[v] - trfy ) * ( By[v] - trfy ) );
							if( diff < mindiff )
							{
								mindiff = diff;
								mindiffidx = v;
							}
						}
					}

					if( mindiff < 10.0 )
					{
						if ( pairsB_temp[ mindiffidx ] == -1 ) 
						{
							pairsA_temp[ itrf ] = mindiffidx;
							pairsB_temp[ mindiffidx ] = itrf;
							pairsA_diff[ itrf ] = mindiff;
							valid ++;
							mean += mindiff;
						}
						else
						{
							pairsA_temp[ itrf ] = -1;
							invalid ++;
						}
					}					
					else
					{
						invalid ++;
						pairsA_temp[ itrf ] = -1;
					}

				}
			}
			
			if( ( novalid == 1 ) || ( valid > maxvalid ) || ( ( valid == maxvalid ) && ( mean < min_mean ) )  )
			{
				novalid = 0;
				maxvalid = valid;
				for( j = 0; j < numA; j++ )
					pairsA[ j ] = pairsA_temp[ j ];

				max_i1 = i1;
				max_i2 = i2;
				max_i3 = i3;
				max_t11 = t11;
				max_t12 = t12;
				max_t13 = t13;
				max_t21 = t21;
				max_t22 = t22;
				max_t23 = t23;

				min_mean = mean;
			}

			count ++;
		}


		String str = new String();

		for( j = 0; j < numA; j++ )
			str += ( pairsA[j] + 1 ) + " ";

		str += "( " + count + " ) " + ( maxvalid + 3 ) + " " + min_mean;

		IJ.showMessage( str );
	}

}
