import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import ij.plugin.*;

public class Point_Based_Affine implements PlugIn {

//	private ImagePlus imp;
	ImagePlus[] image = new ImagePlus [2];
	ImageProcessor[] imp = new ImageProcessor [2];

	int refnop, movnop;
	int[] refxcoords, refycoords, movxcoords, movycoords;
	int refbx, refby, movbx, movby;

	double t11, t12, t13, t21, t22, t23;

	int width;
	int height;
	byte[] moving_pixels;
	byte[] out_pixels;

	public class Simple_Matrix {
		public int rows;
		public int columns;

		double[] mtx;

		Simple_Matrix( int r, int c )
		{
			mtx = new double [ r * c ];
			rows = r;
			columns = c;
		}

		double getElement(int r, int c)
		{
			return mtx [ r * columns + c ];
		}
		
		void setElement(int r, int c, double value)
		{
			mtx [ r * columns + c ] = value;
		}

		void show()
		{
			String str = new String();
			
			for( int i = 0; i < rows; i++ )
			{
				for( int j = 0; j < columns; j++ )
				{
					str += getElement( i, j );
					str += " ";
				}

				str += "\n";
			}

			IJ.showMessage( str );
		}
		
	}

	private int Jordan ( Simple_Matrix s )
	{
		int i, j, k;
		double x;
		double value;

		for(i=0; i<s.rows; i++)
		{
//			s.show3x4();

			x = s.getElement( i, i );
			if( x == 0.0)
			{
				for( j = i + 1; j < s.rows; j++ )
				if(s.getElement( i, i ) != 0.0)
					break;

				if ( j == s.rows )
					return 0;

			for( k = 0; k < s.columns; k++)
			{
				x = s.getElement( i, k );
				value = s.getElement( j, k );
				s.setElement( i, k, value );
				s.setElement( j, k, x );
			}

			x = s.getElement(i, i);
		}

//		IJ.showMessage( "Foelem " + x );

		s.setElement( i, i, 1.0 );
		for( j = i + 1; j < s.columns; j++)
		{
			value = s.getElement( i, j ) / x;
			s.setElement( i, j, value );
//		IJ.showMessage( "Osztas utan "+ value );

		}

		for( j = 0; j < s.rows; j++)
		{
			if( j != i)
			{
				x = s.getElement( j, i );
				s.setElement( j, i, 0.0 );
				for( k = i + 1; k < s.columns; k++)
				{
					value = s.getElement( j, k ) - x * s.getElement( i, k );
					s.setElement( j, k, value );
				}
			}
		}
	}



		return 1;
	}

	public void run(String arg) {
		int i, j;
		int x, y;
		int xtrf, ytrf;
		int w2, h2;

		if( imagesDialog() == false )
			return;

		if( typeCheck() == false )
			return;

		if ( getCoordinates() == false )
			return;

		solveProblem();

//

		width = imp[0].getWidth();
		height = imp[0].getHeight();

		w2 = imp[1].getWidth();
		h2 = imp[1].getHeight();

		moving_pixels = (byte[]) imp[1].getPixels();

		// create a new image with the same size and copy the pixels of the original image
		ImagePlus transformed = NewImage.createByteImage ("Transformed", width, height, 1, NewImage.FILL_BLACK);
		ImageProcessor transformed_ip = transformed.getProcessor();
		out_pixels = (byte[]) transformed_ip.getPixels();

//
		int pos1, pos2;
		for( y = 0; y < height; y++ )
			for( x = 0; x < width; x++ )
			{
				xtrf = (int) Math.round( t11 * x + t12 * y + t13 );
				ytrf = (int) Math.round( t21 * x + t22 * y + t23 );

				if( ( xtrf >= 0 ) && ( xtrf < w2 ) && ( ytrf >= 0 ) && ( ytrf < h2 ) )
				{
					pos1 = y * width + x;
					pos2 = ytrf * w2 + xtrf;

					out_pixels[ pos1 ] = moving_pixels[ pos2 ];
				} 
			}

//
		transformed.show();
		transformed.updateAndDraw();
	}

	public boolean imagesDialog()
	{
		int[] wList = WindowManager.getIDList();
		if ( ( wList==null ) || ( wList.length < 2 ) ) {
			IJ.error("At least two images required.");
			return false;
		}

		String[] titles = new String[wList.length];
		for (int i=0; i<wList.length; i++) {
			ImagePlus imp = WindowManager.getImage(wList[i]);
			titles[i] = imp!=null?imp.getTitle():"";
		}

		GenericDialog gd = new GenericDialog("PointSelection Test");
		gd.addChoice("Reference:", titles, titles[0]);
		gd.addChoice("Moving:", titles, titles[1]);

		gd.showDialog();
		if (gd.wasCanceled()) 
			return false;
		for(int ii=0;ii<2;ii++) {
			int FILE=gd.getNextChoiceIndex();
			image[ii] = (FILE<wList.length)
					? WindowManager.getImage(wList[FILE])
					: null;

			imp[ii] = ( image[ii] != null)
					? image[ii].getProcessor()
					: null;
		}

		return true;
	}

	private boolean typeCheck()
	{
		int type;

		type = image[0].getType();

		if ( type != ImagePlus.GRAY8 )
		{
			IJ.error("Reference image must be 8-bit grayscale");
			return false;
		}

		type = image[1].getType();

		if ( type != ImagePlus.GRAY8 )
		{
			IJ.error("Moving image must be 8-bit template grayscale");
			return false;
		}

		return true;
	}

	private boolean getCoordinates()
	{
		Roi refroi = image[1].getRoi();
		int roiType = refroi.getType();

		if ( roiType != 10 )
		{
			IJ.error("Point selection required!");
			return false;
		}

		Rectangle refrect = refroi.getBounds();
		refbx = refrect.x;
		refby = refrect.y;
//		IJ.showMessage( "refbx, refby: " + refbx + " " + refby );

		PolygonRoi refp = ( PolygonRoi ) refroi;
	            refnop = refp.getNCoordinates();
	            refxcoords = refp.getXCoordinates();
	            refycoords = refp.getYCoordinates();

//			IJ.showMessage( "refnop: " + refnop );			
//		IJ.showMessage( "(" + refxcoords[0] + ", " + refycoords[0] + ")" );
//		IJ.showMessage( "(" + refxcoords[1] + ", " + refycoords[1] + ")" );


//

		Roi movroi = image[0].getRoi();
		roiType = movroi.getType();

		if ( roiType != 10 )
		{
			IJ.error("Point selection required!");
			return false;
		}

		Rectangle movrect = movroi.getBounds();
		movbx = movrect.x;
		movby = movrect.y;

		PolygonRoi movp = ( PolygonRoi ) movroi;
	            movnop = movp.getNCoordinates();
	            movxcoords = movp.getXCoordinates();
	            movycoords = movp.getYCoordinates();

//		IJ.showMessage( "(" + movxcoords[0] + ", " + movycoords[0] + ")" );
//		IJ.showMessage( "(" + movxcoords[1] + ", " + movycoords[1] + ")" );



//			IJ.showMessage( "movnop: " + movnop );			

		if ( refnop != movnop )
		{
			IJ.error( "Number of selected points must be the same! " );
			return false;
		}

		return true;
	}

	private boolean solveProblem()
	{
		Simple_Matrix s = new Simple_Matrix( 3, 4 );

		s.setElement( 0, 0, movbx + movxcoords[0] );
		s.setElement( 0, 1, movby + movycoords[0] );
		s.setElement( 0, 2, 1.0 );
		s.setElement( 0, 3, refbx + refxcoords[0] );

		s.setElement( 1, 0, movbx + movxcoords[1] );
		s.setElement( 1, 1, movby + movycoords[1] );
		s.setElement( 1, 2, 1.0 );
		s.setElement( 1, 3, refbx + refxcoords[1] );

		s.setElement( 2, 0, movbx + movxcoords[2] );
		s.setElement( 2, 1, movby + movycoords[2] );
		s.setElement( 2, 2, 1.0 );
		s.setElement( 2, 3, refbx + refxcoords[2] );

//		s.show();

		Jordan( s );

		t11 = s.getElement( 0, 3 );
		t12 = s.getElement( 1, 3 );
		t13 = s.getElement( 2, 3 );

//

		s.setElement( 0, 0, movbx + movxcoords[0] );
		s.setElement( 0, 1, movby + movycoords[0] );
		s.setElement( 0, 2, 1.0 );
		s.setElement( 0, 3, refby + refycoords[0] );

		s.setElement( 1, 0, movbx + movxcoords[1] );
		s.setElement( 1, 1, movby + movycoords[1] );
		s.setElement( 1, 2, 1.0 );
		s.setElement( 1, 3, refby + refycoords[1] );

		s.setElement( 2, 0, movbx + movxcoords[2] );
		s.setElement( 2, 1, movby + movycoords[2] );
		s.setElement( 2, 2, 1.0 );
		s.setElement( 2, 3, refby + refycoords[2] );

//		s.show();

		Jordan( s );

		t21 = s.getElement( 0, 3 );
		t22 = s.getElement( 1, 3 );
		t23 = s.getElement( 2, 3 );

//		IJ.showMessage( t11 + " " + t12 + " " + t13 + "\n" + t21 + " " + t22 + " " + t23 );

		return true;
	}

}
